##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import zope.interface
import zope.schema


class IOIDAware(zope.interface.Interface):
    """OID aware object."""

    oid = zope.schema.Int(
        title=u"OID",
        description=u"The object id",
        default=None,
        readonly=False,
        required=False)


class IOIDManager(zope.interface.Interface):
    """Object id management adapter.
    
    This adapter can adapt any persistent object and offers access to the 
    objects located in the same database where the adapted object is stored.
    """

    def add(obj):
        """Add an object will activate the access to the object by apply an oid
        """

    def remove(obj):
        """Remove an object will deactivate the access to the object by remove
        an oid.
        """

    def validate(oid):
        """Returns True if an object is valid if not return False."""

    def getObject(oid):
        """Returns an object by the given object id.
        
        Raises a KeyError if no object can be found e.g. removed objects.
        """

    def queryObject(oid, default=None):
        """Return an object if availble if not return default."""

    def getOID(obj):
        """Returns an object id for the given object.
        
        Raises a KeyError if no object can be found e.g. non persistent objects.
        """

    def queryOID(obj, default=None):
        """Returns an oid if available or default if not."""



# events
class IOIDEvent(zope.interface.Interface):
    """Generic base interface for events"""

    object = zope.interface.Attribute("The object related to this event")


class IOIDRemovedEvent(IOIDEvent):
    """Before a oid will be removed from the object

    The event get notified BEFORE the oid is removed from the object.
    """


class IOIDAddedEvent(IOIDEvent):
    """After a unique oid get added to the object

    The event get notified AFTER an oid get added to the object.
    """



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import persistent.interfaces
from ZODB.utils import p64
from ZODB.utils import u64
from ZODB.POSException import ConnectionStateError
from ZODB.POSException import POSKeyError

import zope.interface
import zope.component
from zope.schema.fieldproperty import FieldProperty

import p01.oid.api
from p01.oid import interfaces


class OIDAware(object):
    """OID aware mixin class."""

    zope.interface.implements(interfaces.IOIDAware)
    
    _oid = FieldProperty(interfaces.IOIDAware['oid'])

    @property
    def oid(self):
        return self._oid


class OIDManager(object):
    """Object id management adapter.
    
    This adapter can adapt any persistent object and offers access to the 
    objects located in the same database where the adapted object is stored.
    """

    zope.interface.implements(interfaces.IOIDManager)
    zope.component.adapts(persistent.interfaces.IPersistent)

    def __init__(self, context):
        self.context = context
        self.conn = context._p_jar or p01.oid.api.queryConnection(context)

    def add(self, obj):
        """Add an object will activate the access to the object by apply an oid
        """
        if obj.oid is None:
            if obj._p_oid is None:
                # if _p_oid is None, add the obj to the connection
                self.conn.add(obj)
            obj._oid = u64(obj._p_oid)
        return obj.oid

    def remove(self, obj):
        """Remove an object will deactivate the access to the object by remove
        an oid.
        """
        obj._oid = None

    def validate(self, oid):
        """Returns True if an object is valid if not return False."""
        obj = self.conn.get(p64(oid))
        if obj.oid == oid:
            return True
        return False

    def getObject(self, oid):
        """Returns an object by the given object id."""
        try:
            obj = self.conn.get(p64(oid))
            if obj.oid is not None:
                return obj
        except (ConnectionStateError, POSKeyError, AttributeError), e:
            pass
        raise KeyError(oid)

    def queryObject(self, oid, default=None):
        """Return an object if availble if not return default."""
        try:
            return self.getObject(oid)
        except KeyError, e:
            pass
        return default

    def getOID(self, obj):
        """Returns an object id for the given object."""
        return obj.oid

    def queryOID(self, obj, default=None):
        """Returns an oid if available or default if not."""
        if obj.oid is not None:
            return obj.oid
        return default

    def __repr__(self):
        return '<%s for %s>' %(self.__class__.__name__, self.context)



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import doctest
import unittest
from zope.testing.doctestunit import DocFileSuite

from p01.oid import testing


def test_suite():
    return unittest.TestSuite((
        DocFileSuite('README.txt',
            setUp=testing.setUpOID,
            tearDown=testing.tearDownOID,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS),
        ))


if __name__ == '__main__':
    unittest.main()


##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

from ZODB.utils import u64

import zope.interface
import zope.component
import zope.component.hooks
import zope.lifecycleevent.interfaces

from p01.oid import interfaces


# events
class OIDRemovedEvent(object):
    """Before a unique oid will be removed from the object

    The event get notified BEFORE the oid get set to NONE. This will allow to
    unindex objects be their oid in catalog indexes beforethe object get
    unaccessible by it's oid.
    """

    zope.interface.implements(interfaces.IOIDRemovedEvent)

    def __init__(self, object):
        self.object = object


class OIDAddedEvent(object):
    """After a unique oid get added to the object

    The event get notified AFTER an oid get applied.
    """

    zope.interface.implements(interfaces.IOIDAddedEvent)

    def __init__(self, object):
        self.object = object


# subscriber
@zope.component.adapter(interfaces.IOIDAware,
                        zope.lifecycleevent.interfaces.IObjectCreatedEvent)
def addOIDSubscriber(obj, event):
    """A subscriber to IObjectCreatedEvent

    Add an _oid based on the _p_oid which will enable access to the object in
    the database
    """
    if obj.oid is None:
        if obj._p_oid is None:
            # add the object to the connection
            site = zope.component.hooks.getSite()
            conn = site._p_jar
            conn.add(obj)
        obj._oid = u64(obj._p_oid)


@zope.component.adapter(interfaces.IOIDAware,
                        zope.lifecycleevent.interfaces.IObjectRemovedEvent)
def removeOIDSubscriber(obj, event):
    """A subscriber to IObjectRemovedEvent

    Removes the _oid which will disable access to the object in the database
    """
    obj._oid = None


@zope.component.adapter(interfaces.IOIDEvent)
def dispatchIOIDEvents(event):
    """Event subscriber to dispatch IOIDEvent to interested adapters."""
    for adapted in zope.component.subscribers((event.object, event), None):
        pass # adapting them does the work



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup

$Id:$
"""
import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup (
    name='p01.oid',
    version='0.5.0',
    author = "Roger Ineichen, Projekt01 GmbH",
    author_email = "dev@projekt01.ch",
    description =
        "Conflict free unique ID concept for Zope3 components (with ZODB)",
    long_description=(
        read('README.txt')
        + '\n\n' +
        read('CHANGES.txt')
        ),
    license = "ZPL 2.1",
    keywords = "zope z3c p01 unique id utility components Zope3",
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope3'],
    url = 'http://pypi.python.org/pypi/p01.oid',
    packages = find_packages('src'),
    include_package_data = True,
    package_dir = {'':'src'},
    namespace_packages = ['p01'],
    extras_require=dict(
        test=[
            'zope.testing',
            'zope.site',
            ],
        ),
    install_requires = [
        'setuptools',
        'ZODB3',
        'zope.component',
        'zope.interface',
        'zope.lifecycleevent',
        'zope.schema',
        ],
    zip_safe = False,
    )



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

from ZODB.utils import p64
from ZODB.utils import u64
from ZODB.POSException import ConnectionStateError
from ZODB.POSException import POSKeyError

from zope.component.hooks import getSite


def _queryCon(cur):
    if cur._p_jar is not None:
        return cur._p_jar
    while not getattr(cur, '_p_jar', None):
        cur = getattr(cur, '__parent__', None)
        if cur is None:
            return None
    return cur._p_jar


def queryConnection(obj, context=None):
    """Returns a connection based on obj, context or site."""
    conn = _queryCon(obj)
    if conn is None and context is not None:
        conn = _queryCon(context)
    if conn is None:
        site = getSite()
        conn = site._p_jar
    return conn


def addObject(obj, context=None):
    if obj.oid is None:
        if obj._p_oid is None:
            # if _p_oid is None, try to add the obj to the connection
            conn = queryConnection(obj, context)
            conn.add(obj)
        obj._oid = u64(obj._p_oid)


def removeObject(obj):
    obj._oid = None


def getObject(oid, context=None):
    if context is None:
        context = getSite()
    conn = queryConnection(context)
    try:
        obj = conn.get(p64(oid))
        if getattr(obj, 'oid', None) is not None:
            return obj
    except (ConnectionStateError, POSKeyError), e:
        pass
    raise KeyError(oid)


def queryObject(oid, context=None, default=None):
    if oid is None:
        return None
    try:
        return getObject(oid, context)
    except KeyError, e:
        pass
    return default


def yieldObjects(oids, context=None):
    """Optimized yield object method.
    
    Only pass a context as argument if the context is available in the 
    connection otherwise an ISite is used as context.
    """ 
    if context is None:
        context = getSite()
    conn = context._p_jar
    for oid in oids:
        try:
            obj = conn.get(p64(oid))
            if getattr(obj, 'oid', None) is not None:
                yield obj
        except (ConnectionStateError, POSKeyError), e:
            pass



##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import persistent
import zope.component.testing
import zope.component.hooks
from zope.site import folder

from p01.oid import oid


class MySite(folder.Folder):
    """Sample site."""

    def __repr__(self):
        return '<%s>' % self.__class__.__name__


class MyObject(persistent.Persistent, oid.OIDAware):
    """Sample object."""

    __name__ = __parent__ = None

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

def setUpOID(test):
    zope.component.hooks.setHooks()


def tearDownOID(test):
    zope.component.hooks.resetHooks()
    zope.component.testing.tearDown()
